/*
 * Copyright (c) 2004 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

/**
 ** aspectExecTaskCodels.c
 **
 ** Codels appeles par la tache aspectExecTask
 **
 ** Auteur: 
 ** Date: 
 **
 **/

#ifndef UNIX
#include <vxWorks.h>
#else
#include <portLib.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include "h2mathLib.h"

#include "libeuler/pomEuler.h"

/* From the sick module */
#include <server/sickPosterLib.h>
#include <sick_convert.h>

/* From mapkit/segkit */
#include <aspectLib.h>

#include "server/aspectHeader.h"
#include "server/aspectPosterReadLibProto.h"
#include "aspectConst.h"
#include "segkitIO.h"

#define ASPECT_DATE int

/* ----------------------------- PROTOS ---------------------------------- */

static BOOL aspectFuseSick(double thMin, double thMax, double dTh, 
			   SICK_SCANPOLAR_POSTER_STR *aspectScan);
static int aspectReallocArrays(int nbMapSegs, int *nbMaxDynSegs);
static int aspectFreeArrays (void);
#if defined DEBUG && 0
static BOOL aspectExportInputMap(int nbSegs, int date, 
				  ASPECT_SEG *segs);
#endif
static BOOL aspectAddBackSegsToInputMap(ASPECT_CARTESIANASPECT_POSTER_STR
					*aspectMap,
					POM_SENSOR_POS *newSickPomPos);

static BOOL testFreshData(unsigned int date, int *bilan);


/* ---------------------------- STATIC VAR ------------------------------ */
static POSTER_ID sickPosterId;

static int nbInputMapSegs;          /* nb segs from map */
static int nbInputSickSegs;         /* nb segs from sick */
static int nbInputDynSegs;          /* nb segs from sick + aspect */
static int nbInputSegs;             /* nb segs currently in the input map */
static int nbMaxInputDynSegs;       /* nb max segs from sick */
static int nbMaxInputSegs;          /* input map capacity */

static int nbAspectSegs;            /* nb segs currently in the aspect map */

static unsigned int sickAcqTime;    /* */

typedef struct ASPECT_INPUT_MAP {
  ASPECT_SEG *segs;     /* all the input segs */
  int date[ASPECT_MAX_ASPECT_MAP_SIZE];   /* dates only for segs coming from 
					     previous aspect map */
  ASPECT_SEG_TYPE type[ASPECT_MAX_ASPECT_MAP_SIZE]; /* types only for segs coming 
						       from previous aspect map */
} ASPECT_INPUT_MAP;
static ASPECT_INPUT_MAP aspectInputMap;

static int aspectSegDate[ASPECT_MAX_ASPECT_MAP_SIZE]; /* to store the seg dates after 
							 aspect computation */

static SEGLOC_MAP mapFromPoster;
static SICK_SEG *mapFromFile;

static ASPECT_CARTESIANASPECT_POSTER_STR aspectCartesianMap;
static ASPECT_POLARASPECT_POSTER_STR aspectPolarMap;

static SICK_SCANPOLAR_POSTER_STR aspectScan;

/*------------------------------------------------------------------------
 *
 * aspectExecTaskInit  --  Codel d'initialisation (SDI/f, ...)
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS aspectExecTaskInit(int *bilan)
{
  SDI_F->exportLaser.simulData = FALSE;
  SDI_F->exportLaser.fuseRealData = FALSE;
  SDI_F->exportLaser.thMinDeg = -90;
  SDI_F->exportLaser.thMaxDeg = 90;
  SDI_F->exportLaser.dThDeg = 0.5;

  SDI_F->viewParameters.maxDistView = ASPECT_DEFAULT_DIST_VIEW_MAX;
  SDI_F->viewParameters.halfAngleView = ASPECT_DEFAULT_HALF_ANGLE_VIEW_MAX;
  SDI_F->viewParameters.keepReversedSegs = ASPECT_DEFAULT_KEEP_REVERSED_SEGS;
  SDI_F->viewParameters.minAspectSegAngle = ASPECT_DEFAULT_MIN_ASPECT_SEG_ANGLE;

  SDI_F->mapInput.type = ASPECT_DEFAULT_MAP_TYPE;
  /*strcpy(SDI_F->mapInput.name, ASPECT_DEFAULT_MAP_SOURCE);*/
  strcpy(SDI_F->segsPosterName.name, ASPECT_DEFAULT_SEGS_SOURCE);

  aspectCartesianMap.Header.frame = SICK_SICK_FRAME;

  nbMaxInputDynSegs = ASPECT_MAX_ASPECT_MAP_SIZE + SICK_MAX_ACQ_SEGS;

  return OK;
}

/*------------------------------------------------------------------------
 *
 * aspectExecTaskEnd  --  Codel de terminaison
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS aspectExecTaskEnd(void)
{
  if (mapFromFile)
    free(mapFromFile);

  aspectFreeArrays();  
  return OK;
}

/*------------------------------------------------------------------------
 * ReadMap
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_aspect_CANNOT_READ_MAP
 */

/* aspectReadMapExec  -  codel EXEC de ReadMap
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT aspectReadMapExec (ASPECT_INPUT_SOURCE *input, int *nbsegs,
				 int *bilan)
{
  POSTER_ID posterId;

  /* read by file */
  if (input->type == ASPECT_FILE) {

    /* the map segs is allocated by readSegkitMap */
    if (mapFromFile)
      free(mapFromFile);

    if (!(SDI_F->nbReadMapSegs = readSegkitMap(input->name, 0,
					(SICK_SEG **)&mapFromFile))) {
      *bilan = S_aspect_CANNOT_READ_MAP;
      return ETHER;
    }
  }

  /* read by poster */
  else {
    if (posterFind(input->name, &posterId) == ERROR) {
      *bilan = S_aspect_CANNOT_READ_MAP;
      return ETHER;
    }
    if (aspectSEGLOC_MAPPosterRead(posterId, &mapFromPoster) == ERROR) {
      *bilan = S_aspect_CANNOT_READ_POSTER;
      return ETHER;
    }
    SDI_F->nbReadMapSegs = mapFromPoster.header.nbSegs;
  }

  *nbsegs = SDI_F->nbReadMapSegs;
  printf ("aspectReadMap : %d segs in map\n", SDI_F->nbReadMapSegs);

  return ETHER;
}


#if defined DEBUG && 0
/*------------------------------------------------------------------------
 * aspectExportInputMap
 *
 * Description: 
 *
 */

static BOOL aspectExportInputMap(int nbSegs, int date, ASPECT_SEG *segs)
{
  static ASPECT_INPUTMAP_POSTER_STR inputMap;

  /* fill in header */
  inputMap.Aspect_input_map_header.date = date;

  /* test poster size */
  if (nbSegs > ASPECT_MAX_INPUT_MAP_SIZE) {
    printf ("aspectExportInputMap: WARNING poster too small %d segs recorded instead of %d", ASPECT_MAX_INPUT_MAP_SIZE, nbSegs);
    inputMap.Aspect_input_map_header.nbSegs = ASPECT_MAX_INPUT_MAP_SIZE;
  }
  else {
    inputMap.Aspect_input_map_header.nbSegs = nbSegs;
  }

  /* copy segs in structure for the poster :-( XXXX */
  memcpy(inputMap.Aspect_input_map_segs.segs, segs, 
	 sizeof(SICK_SEG) * inputMap.Aspect_input_map_header.nbSegs);
  
  /* write poster */
  if (aspectInputMapPosterWrite(ASPECT_INPUTMAP_POSTER_ID, &inputMap) == ERROR)
    return FALSE;
    
  return TRUE;
}
#endif /* DEBUG */

/*------------------------------------------------------------------------
 * initRecordSickSegs
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_aspect_CANNOT_READ_POSTER
 */

static POSTER_ID segsPosterId=NULL;


/* initRecordSickSegs  -  codel START de DynamicMapBuilding
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
static BOOL initRecordSickSegs (ASPECT_STRING *segsPosterName,
				int *bilan)
{
  /* look for poster */
  if (!segsPosterId) {
    aspectSetMapExec(segsPosterName, bilan);
    if (*bilan != OK) {
      return FALSE;
    }
  }

  /* (re)alloc seg tabs if required */
  if (!aspectReallocArrays(SDI_F->nbReadMapSegs, &nbMaxInputDynSegs)) {
    *bilan = S_aspect_NOT_ENOUGH_MEMORY;
    return FALSE;
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
 *  recordSickSegs  -  
 *
 *  Retourne:  BOOL 
 */
BOOL recordSickSegs (ASPECT_STRING *segsPosterName, int *bilan)
{
  static SICK_SEGS_POSTER_STR sickSegs;

  /* read sick segs */
  if (aspectSICK_SEGS_POSTER_STRPosterRead(segsPosterId, &sickSegs)) {
    *bilan = S_aspect_CANNOT_READ_POSTER;
    segsPosterId = NULL;
    return ETHER;
  }

  if (!testFreshData(sickSegs.Header.sickPomPos.date, bilan)) {
    segsPosterId = NULL;
    return ETHER;
  }

  /* project to global frame AND record in input map */
  sickConvertSegsFrame(&sickSegs.Header, 
		       SICK_GLOBAL_FRAME,
		       sickSegs.Segs.segs,
		       (SICK_SEG *) &aspectInputMap.segs[nbInputSegs]);

  /* update number of segments */
  nbInputSickSegs = sickSegs.Header.nbSegs;
  nbInputDynSegs += nbInputSickSegs;
  nbInputSegs += nbInputSickSegs;
  sickAcqTime = sickSegs.Header.sickTime;

  return EXEC;
}

static BOOL testFreshData(unsigned int date, int *bilan)
{
  static unsigned int prevDate=-1;
  static int count=0;

  if (date == prevDate) {
/*       printf("***** aspect : sick not updated %d times (date = %d)\n", count, date); */
    count++;
  }
  else
    count = 0;
  if (count > 20) {
    *bilan = S_aspect_CANNOT_READ_POSTER;
    count = 0;
    fprintf(stderr, "aspect : sickPoster not fresh (date %d)\n", 
	    date);
/*     return FALSE; */
  return TRUE;  }
  prevDate = date;
  return TRUE;
}


/*------------------------------------------------------------------------
 * ResetMap
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_aspect_CANNOT_READ_POSTER
 */

/* aspectResetMapExec  -  codel EXEC de ResetMap
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT aspectResetMapExec (int *bilan)
{
  if (nbInputSegs < nbInputDynSegs) {
    printf ("aspectResetMapExec: nbInputSegs = %d < nbInputDynSegs = %d ?\n", 
	    nbInputSegs, nbInputDynSegs);
  }
  
  nbInputSegs = nbInputMapSegs;
  nbInputDynSegs = 0;
  aspectCartesianMap.Header.nbSegs = 0;
  return ETHER;
}


/*------------------------------------------------------------------------
 * LoadMap
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_aspect_CANNOT_READ_POSTER
 */
static int loadMap(ASPECT_INPUT_SOURCE *inputMap, int *bilan)
{
  /* read map if not done */
  if (!SDI_F->nbReadMapSegs) {
    aspectReadMapExec (inputMap, &SDI_F->nbReadMapSegs, bilan);
    if (*bilan != OK)
      return 0;
  }

  /* alloc seg tabs */
  if (!aspectReallocArrays(SDI_F->nbReadMapSegs, &nbMaxInputDynSegs)) {
    *bilan = S_aspect_NOT_ENOUGH_MEMORY;
    return ETHER;
  }
 
  /* copy segs in inputMap */
  if (inputMap->type == ASPECT_FILE)
    memcpy(aspectInputMap.segs, mapFromFile,   
	   SDI_F->nbReadMapSegs*sizeof(SICK_SEG));
  else 
    memcpy(aspectInputMap.segs, mapFromPoster.segs, 
	   SDI_F->nbReadMapSegs*sizeof(SICK_SEG));

  /* remove previous map segs if any */
  nbInputSegs = nbInputDynSegs;

  /* record new map segs */
  nbInputMapSegs = SDI_F->nbReadMapSegs;
  nbInputSegs += nbInputMapSegs;

  return nbInputMapSegs;
}

static POSTER_ID posPosterId;
/*------------------------------------------------------------------------
 * AspectFromPosterConfig
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_aspect_CANNOT_READ_POSTER
 */

/* aspectAspectFromPosterConfigStart  -  codel EXEC de AspectFromPosterConfig
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT aspectAspectFromPosterConfigStart (ASPECT_INPUTS_POSTER *inputsPoster, 
						  int *bilan)
{
  static char posterName[64];

  aspectResetMapExec(bilan);

  /* load map if required */
  if (inputsPoster->inputFrom.mapSegs) {
    if (!(nbInputMapSegs = loadMap(&SDI_F->mapInput, bilan)))
      return ETHER;
  }
  else {
    nbInputMapSegs = 0;
  }

  /* look for sick segs if required */
  if (inputsPoster->inputFrom.sickSegs) {
    if (!initRecordSickSegs(&SDI_F->segsPosterName, bilan))
      return ETHER;
    if (!recordSickSegs(&SDI_F->segsPosterName, bilan))
      return ETHER;
  }
  else
    nbInputSickSegs = 0;

  /* test map valid */
  if (!nbInputSegs) {
    printf ("aspectAspectFromPosterConfigStart: empty map\n");
    *bilan = S_aspect_EMPTY_MAP;
    return ETHER;
  }

  /* look for poster config */
  if (strcmp(posterName, inputsPoster->posPosterName.name))
    posPosterId = NULL;

  if (!posPosterId) {
    if (posterFind(inputsPoster->posPosterName.name, &posPosterId) == ERROR) {
      *bilan = S_aspect_CANNOT_READ_POSTER;
      return ETHER;
    }
    strcpy(posterName, inputsPoster->posPosterName.name);
  }

  return EXEC;
}

/* aspectAspectFromPosterConfigExec  -  codel EXEC de AspectFromPosterConfig
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT aspectAspectFromPosterConfigExec (ASPECT_INPUTS_POSTER *inputsPoster, 
						 int *bilan)
{
  ASPECT_CONFIG sensorConfig;
  int nbBackSegs;
  POM_SENSOR_POS newSickPomPos;
  static POM_EULER_V t3d;
  int segsid[ASPECT_MAX_ASPECT_MAP_SIZE];
  int i, oldAspectSegId;

  /**
   ** rebuild the input map 
   **/
  nbInputSegs = nbInputMapSegs;
  nbInputDynSegs = 0;

  /* look for sick segs if required */
  if (inputsPoster->inputFrom.sickSegs) {
    if (!recordSickSegs(&SDI_F->segsPosterName, bilan))
      return ETHER;
  }
  else 
    nbInputSickSegs = 0;

  /* read current SICK POS */
  if (aspectPOM_SENSOR_POSPosterRead(posPosterId, &newSickPomPos) == ERROR) {
    posPosterId = NULL;
    *bilan = S_aspect_CANNOT_READ_POSTER;
    return ETHER;
  }

  /* computes sick pos in global frame */
  pomEulerVCompIn(&t3d, &newSickPomPos.sensorToMain, 
		  &newSickPomPos.mainToOrigin);
  sensorConfig.x = t3d.euler.x;
  sensorConfig.y = t3d.euler.y;
  sensorConfig.theta = t3d.euler.yaw;

  /* keep aspect segs that are in the back (ie, not visible) */
  if (inputsPoster->inputFrom.keepSickSegsInBack) {
    if (!(nbBackSegs = aspectAddBackSegsToInputMap(&aspectCartesianMap,
						   &newSickPomPos))) {
      *bilan = errnoGet();
      return ETHER;
    }
  }

  /* test map validity */
  if (!nbInputSegs) {
    printf ("aspectAspectFromPosterConfigStart: empty map\n");
    *bilan = S_aspect_EMPTY_MAP;
    return ETHER;
  }

  /**
   ** main computation 
   **/
  nbAspectSegs = aspectMapFillIn(&sensorConfig, 
				 &SDI_F->viewParameters,
				 /* input map */
				 nbInputSegs, 
				 aspectInputMap.segs, 
				 FALSE, /* results in local frame */
				 /* where to fill result */
				 SICK_MAX_ACQ_SEGS,
				 (ASPECT_SEG *)aspectCartesianMap.Segs.segs,
				 aspectPolarMap.PolarSegs.segs,
				 segsid,
				 /* simul laser */
				 0, NULL);

#ifdef DEBUG
  printf ("%2d map + %2d sick + %2d back aspect = %2d inputs -> %2d output\n", 
	  nbInputMapSegs,
	  nbInputSickSegs,
	  nbBackSegs,
	  nbInputSegs,
	  nbAspectSegs);
#endif

  /* Fill in poster header */
  aspectCartesianMap.Header.nbSegs = nbAspectSegs;
  aspectCartesianMap.Header.nbLonelyPoints = 0;
  aspectCartesianMap.Header.frame = SICK_SICK_FRAME;
  aspectCartesianMap.Header.sickPomPos = newSickPomPos;
  aspectCartesianMap.Header.sickTime = sickAcqTime;
  aspectPolarMap.Header = aspectCartesianMap.Header;

   /* laser simulation */
  if (SDI_F->exportLaser.simulData) {
     aspectSimulLaser(DEG_TO_RAD(SDI_F->exportLaser.thMinDeg),
                    DEG_TO_RAD(SDI_F->exportLaser.thMaxDeg),
                    DEG_TO_RAD(SDI_F->exportLaser.dThDeg),
                    nbAspectSegs,
                    aspectPolarMap.PolarSegs.segs,
                    (ASPECT_SEG *)aspectCartesianMap.Segs.segs, 
                    SICK_N_PT_MAX, aspectScan.Rho);
 
     /* fusion with real sick */
     if (SDI_F->exportLaser.fuseRealData)
       aspectFuseSick(DEG_TO_RAD(SDI_F->exportLaser.thMinDeg),
		      DEG_TO_RAD(SDI_F->exportLaser.thMaxDeg),
		      DEG_TO_RAD(SDI_F->exportLaser.dThDeg),
		      &aspectScan);

     /* pos */
     aspectScan.Header.sickPomPos = newSickPomPos;
  }

  /* record segs type and date */
  for (i=0; i<nbAspectSegs; i++) {
    /* map segs */
    if (segsid[i] < nbInputMapSegs) {
      aspectPolarMap.SegsType.segType[i] = ASPECT_MAP_SEG;
      aspectSegDate[i] = 0;
    }
    /* sick segs */
    else if (segsid[i] < nbInputMapSegs + nbInputSickSegs) {
      aspectPolarMap.SegsType.segType[i] = ASPECT_LASER_SEG;
      aspectSegDate[i] = 1;
    }
    /* old aspect segs */
    else {
      oldAspectSegId = segsid[i] - (nbInputMapSegs+nbInputSickSegs);
      aspectPolarMap.SegsType.segType[i] = aspectInputMap.type[oldAspectSegId];
      aspectSegDate[i] = aspectInputMap.date[oldAspectSegId] + 1;
    }
  }

  /* export aspect cartesian map in poster */
  aspectCartesianAspectPosterWrite(ASPECT_CARTESIANASPECT_POSTER_ID, 
				   &aspectCartesianMap);
  /* export aspect polar map in poster */
  aspectPolarAspectPosterWrite(ASPECT_POLARASPECT_POSTER_ID, &aspectPolarMap);
  /* export laser simul */
  if (SDI_F->exportLaser.simulData)
    aspectLaserEchosPosterWrite(ASPECT_LASERECHOS_POSTER_ID,
				(ASPECT_LASERECHOS_POSTER_STR *)&aspectScan);

  /* export input map in poster (XXXX for debug) */
#ifdef DEBUG
  if (!aspectExportInputMap(nbInputSegs, aspectDate, aspectInputMap.segs)) {
    *bilan = errnoGet();
  }
#endif
  return EXEC;
}


/* --------------------------- LOCAL FUNCTIONS ------------------------- */

/* ----------------------------------------------------------------------
 * aspectReallocArrays
 *
 *
 */

static int aspectReallocArrays(int nbMapSegs, int *nbMaxInputDynSegs)
{
  int minRequiredSize = nbMapSegs + *nbMaxInputDynSegs;

  /* already allocated */
  if (aspectInputMap.segs) {
    
    /* -- 
       -- array large enough 
       -- */
    if (nbMaxInputSegs >= minRequiredSize) {

      *nbMaxInputDynSegs = nbMaxInputSegs - nbMapSegs;
      return nbMaxInputSegs;
    }

    /* -- 
       -- arrays too small : realloc 
       -- */
    /* re-allocation segs (to not lose static map if any) */
    if (!(aspectInputMap.segs = realloc(aspectInputMap.segs,
				       minRequiredSize*sizeof(SICK_SEG)))) {
      printf ("aspectReallocArrays: failed realloc %d\n",
	      minRequiredSize*sizeof(SICK_SEG));
      return 0;
    }
    
    nbMaxInputSegs = minRequiredSize;
    return nbMaxInputSegs;
  }
 
  /* --
     -- allocation segs 
     -- */
  if (!(aspectInputMap.segs = malloc(minRequiredSize*sizeof(SICK_SEG)))){
    printf ("aspectReallocArrays: failed alloc %d\n",
	    minRequiredSize*sizeof(SICK_SEG));
    return 0;
  }
  
  /* done */
  nbMaxInputSegs = minRequiredSize;
  return nbMaxInputSegs;
}

static int aspectFreeArrays (void) 
{
  if (aspectInputMap.segs) {
    free(aspectInputMap.segs);
    aspectInputMap.segs = NULL;
  }
  nbMaxInputSegs = 0;
  return 0;
}


/*------------------------------------------------------------------------
 * ExportLaser
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_aspect_POSTER_NOT_FOUND
 */

/* aspectExportLaserExec  -  codel EXEC de ExportLaser
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT aspectExportLaserExec (ASPECT_EXPORT_LASER *exportLaser, 
				      int *bilan)
{
  if (exportLaser->fuseRealData) {
    if (posterFind("sickScanPolar", &sickPosterId) == ERROR) {
      *bilan = S_aspect_POSTER_NOT_FOUND;
      printf ("aspectExportLaserExec : poster sickSegs NOT found\n");
      return ETHER;
    }
    printf ("aspectExportLaserExec : poster sickSegs found\n");
  }
  return ETHER;
}


/*------------------------------------------------------------------------
 * aspectFuseSick
 *
 * Description: 
 *
 */

static BOOL aspectFuseSick(double thMin, double thMax, double dTh, 
			   SICK_SCANPOLAR_POSTER_STR *aspectScan)
{
  int i, firstAspect, aspect;
  static SICK_SCANPOLAR_POSTER_STR sickScan;
  double ratio;

  /* look for poster */
  if (!sickPosterId) {
    if (posterFind("sickScanPolar", &sickPosterId) == ERROR) {
      return FALSE;
    }
  }

  /* lecture poster */
  if (aspectSICK_SCANPOLAR_POSTER_STRPosterRead(sickPosterId, 
						&sickScan) == ERROR) {
    sickPosterId = NULL;
    return FALSE;
  }

  /* look for first apect pt that match first sick pt */
  firstAspect = (int) (angleLimit(sickScan.Header.theta_min - thMin)/dTh +0.5);
  if (firstAspect < 0 || firstAspect > angleLimit(thMax - thMin)/dTh) {
    printf ("aspectFuseSick: aspect and sick not compatible (start)\n");
    return FALSE;
  }
  printf ("aspectFuseSick: first aspect = %d\n", firstAspect);

  /* compute increment */
  ratio = sickScan.Header.dtheta / dTh;

  /* fuse */
  for(i=0; i<SICK_N_PT_MAX; i++) {

    /* get aspect index */
    aspect = firstAspect + (int)(((double)i) * ratio + 0.5);

    /* too far */
    if (thMin + dTh * aspect + 0.001 > thMax)
      break;

    /* keep min dist */
    aspectScan->Rho[aspect] = MIN(aspectScan->Rho[aspect], sickScan.Rho[i]);
  }
  aspectScan->Header = sickScan.Header;
  return TRUE;
}


/*------------------------------------------------------------------------
 * aspectAddBackSegsToInputMap
 *
 *
 */
static BOOL aspectAddBackSegsToInputMap(ASPECT_CARTESIANASPECT_POSTER_STR
					*aspectMap,
					POM_SENSOR_POS *newSickPomPos)
{
  int i;
  SICK_SEG *segm;
  int keepNbSegs = 0;

  /* nothing to add */
  if (!aspectMap->Header.nbSegs)
    return TRUE;

  /* project old aspect map in global frame XXXX ? why these 2 steps ? */
  sickConvertSICK_SEGSFrame((SICK_SEGS_POSTER_STR *)aspectMap, 
			    (SICK_SEGS_POSTER_STR *)aspectMap, 
			    SICK_GLOBAL_FRAME);

  /* project old aspect map in new local frame */
  aspectMap->Header.sickPomPos = *newSickPomPos;
  sickConvertSICK_SEGSFrame((SICK_SEGS_POSTER_STR *)aspectMap, 
			    (SICK_SEGS_POSTER_STR *)aspectMap, 
			    SICK_SICK_FRAME);
  
  /* keep only segs that are backward ... */
  for (i=0; i<aspectMap->Header.nbSegs; i++) {

    segm = &aspectMap->Segs.segs[i];


    /* ... and not from map */
    if (aspectSegDate[i] == 0) {
/*       printf ("aspect: rm map seg %2d (%5.2f %5.2f) (%5.2f %5.2f)\n", */
/* 	      i, segm->x1, segm->y1, segm->x2, segm->y2); */
      continue;
    }

    /* ... and not too old */
    if (aspectSegDate[i] > SDI_F->inputsPoster.inputFrom.backSegsAgeLimit) {
/*       printf ("aspect: rm old seg %2d date %d (%5.2f %5.2f) (%5.2f %5.2f)\n", */
/* 	      i, aspectSegDate[i],  segm->x1, segm->y1, segm->x2, segm->y2); */
      continue;
    }

    /* ... abscisses in sensor frame are in front */
    if (segm->x1 >= 0 && segm->x2 >= 0) {
/*       printf ("aspect: rm front seg %d (%5.2f %5.2f) (%5.2f %5.2f)\n", */
/* 	      i, segm->x1, segm->y1, segm->x2, segm->y2); */
      continue;
    }
    
#if 0
    /* seg extremity 1 cut */
    if (segm->x1 > 0) {
      /* value for x==0 */
      segm->y1 = segm->y2 - 
		 segm->x2 * (segm->y2 - segm->y1)/(segm->x2 - segm->x1);
      segm->x1 = 0;
    }

    /* seg extremity 2 cut */
    else if (segm->x2 > 0) {
      /* value for x==0 */
      segm->y2 = segm->y1 - 
		 segm->x1 * (segm->y2 - segm->y1)/(segm->x2 - segm->x1);
      segm->x2 = 0;
    }
#endif

    aspectMap->Segs.segs[keepNbSegs] = aspectMap->Segs.segs[i];

    aspectInputMap.type[keepNbSegs] = aspectPolarMap.SegsType.segType[i];
    aspectInputMap.date[keepNbSegs] = aspectSegDate[i];

    keepNbSegs++;

#ifdef DEBUG
    printf ("add old seg %d %5.2f %5.2f %5.2f %5.2f\n",
	    i, segm->x1, segm->y1, segm->x2, segm->y2);
#endif
  }

  aspectMap->Header.nbSegs = keepNbSegs;

  /* project back in global frame the remaining segs */
  sickConvertSICK_SEGSFrame((SICK_SEGS_POSTER_STR *)aspectMap, 
			    (SICK_SEGS_POSTER_STR *)aspectMap, 
			    SICK_GLOBAL_FRAME);


  /* add these segs in map */
  if (nbInputSegs + keepNbSegs > nbMaxInputSegs) {
    printf("aspectAddBackSegsToInputMap: too many segs, %d/%d recorded\n", 
	   nbMaxInputSegs - nbInputSegs, keepNbSegs);
    keepNbSegs = nbMaxInputSegs - nbInputSegs;
  }
  memcpy(&aspectInputMap.segs[nbInputSegs], aspectMap->Segs.segs,
	 keepNbSegs * sizeof(SICK_SEG));

  /* record number of segs */
  nbInputDynSegs += keepNbSegs;
  nbInputSegs += keepNbSegs;

  return TRUE;
}

/*------------------------------------------------------------------------
 * SetDynamicSegsSource
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_aspect_CANNOT_READ_MAP
 */

/* aspectSetMapExec  -  codel EXEC de SetDynamicSegsSource
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT aspectSetMapExec (ASPECT_STRING *segsPosterName, int *bilan)
{
  static char posterName[32];

  aspectResetMapExec(bilan);

  /* if new poster, update posterId */
  if (strcmp(posterName, segsPosterName->name))
    segsPosterId = NULL;

  /* look for poster */
  if (!segsPosterId) {
    if (posterFind(segsPosterName->name, &segsPosterId) == ERROR) {
      *bilan = S_aspect_CANNOT_READ_POSTER;
      return FALSE;
    }
    strcpy(posterName, segsPosterName->name);
  }

  return ETHER;
}



