#ifdef UNIX
#include "portLib.h"
#else
#include "vxWorks.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pomStruct.h"
#include "sickStruct.h"
#include "segkitIO.h"

int readSegkitMap(char *filename, int nbMaxSegs, SICK_SEG *mapSegs[])
{
  FILE *pfile;
  int i;
  int nbSegs;
  SICK_SEG *seg;
  double x, y, cs, sn, cst, snt;
  char coosys;
  double enc[4];
  double vx, vy, d,r, o;
  double vr, vro, vo;
  char typ1[8], typ2[8];
  int allocated;

  /* Lecture fichier */
  if ((pfile = fopen(filename,"r")) == NULL) {
    printf ("readSegkitMap: cannot open %s\n", filename);
    return 0;
  }
  
  /* Nombre de segments ? */
  if (fscanf(pfile, "%d", &nbSegs) != 1) {
    printf ("readSegkitMap: cannot read nbsegs\n");
    fclose(pfile);
    return 0;
  }
  if (*mapSegs && nbMaxSegs != 0 && nbMaxSegs<nbSegs) {
    printf ("readSegkitMap: WARNING %d segs in the map for an array of %d segs\n",
	    nbSegs, nbMaxSegs);
    nbSegs = nbMaxSegs;
  }

  /* Entete */
  if (fscanf(pfile, "%lf %lf %lf %lf %lf %lf %c", 
	     &x, &y, 
	     &cs, &sn,
	     &cst, &snt, 
	     &coosys) != 7) {
    printf ("readSegkitMap: cannot read header\n");
    fclose(pfile);
    return 0;
  }
  if (fscanf(pfile, "%lf %lf %lf %lf", 
	     &enc[0], &enc[1],
	     &enc[2], &enc[3]) != 4) {
    printf ("readSegkitMap: cannot read enc\n");
    fclose(pfile);
    return 0;
  }
	     
  /* Les segments */
  allocated = FALSE;
  if (!*mapSegs) {
    if ((*mapSegs = (SICK_SEG *) malloc (sizeof(SICK_SEG) * nbSegs)) == NULL) {
      printf ("readSegkitMap: cannot alloc segs\n");
      fclose(pfile);
      return 0;
    }
    allocated = TRUE;
  }

  seg=*mapSegs;
  for(i=0; i<nbSegs; i++,seg++) {
    
    /* Premier point */
    if (fscanf(pfile, "%lf %lf", 
	       &seg->x1, &seg->y1) != 2) {
      printf ("readSegkitMap: cannot read first point (seg %d)\n", i);
      fclose(pfile);
      if (allocated) free(mapSegs);
      return 0;
    }
    /* second point */
    if (fscanf(pfile, "%lf %lf", 
	       &seg->x2, &seg->y2) != 2) {
      printf ("readSegkitMap: cannot read second point (seg %d)\n", i);
      fclose(pfile);
      if (allocated) free(mapSegs);
      return 0;
    }
    /* vecteur */
    if (fscanf(pfile, "%lf %lf %lf", 
	       &vx, &vy, &d) != 3) {
      printf ("readSegkitMap: cannot read vx ... (seg %d)\n", i);
      fclose(pfile);
      if (allocated) free(mapSegs);
      return 0;
    }
    /* o,r */
    if (fscanf(pfile, "%lf %lf", 
	       &r, &o) != 2) {
      printf ("readSegkitMap: cannot read o, r (seg %d)\n", i);
      fclose(pfile);
      if (allocated) free(mapSegs);
      return 0;
    }
    /* variances */
    if (fscanf(pfile, "%lf %lf %lf", 
	       &vr, &vro, &vo) != 3) {
      printf ("readSegkitMap: cannot read variances ... (seg %d)\n", i);
      fclose(pfile);
      if (allocated) free(mapSegs);
      return 0;
    }
    /* type extremite */
    if (fscanf(pfile, "%s %s", 
	       &typ1, &typ2) != 2) {
      printf ("readSegkitMap: cannot read extremity types (seg %d)\n", 
	      i);
      fclose(pfile);
      if (allocated) free(mapSegs);
      return 0;
    }
  }
    
  fclose(pfile);
  return nbSegs;
}
